import torch
from torch._C import parse_schema
import torch.nn as nn
import torch.nn.functional as F
import torchvision.datasets as dsets
import general_energy_nets.predictive_coding as pc
import torchvision
import torch.optim as optim
import argparse
import matplotlib
import matplotlib.pyplot as plt
import torchvision.transforms as transforms
from tqdm import tqdm
from torchvision import datasets
from torch.utils.data import Dataset, DataLoader, TensorDataset
from torchvision.utils import save_image
from torch.autograd import Variable
import Utils_ass_test_images as u
import statistics
import warnings
import sys
import numpy as np
import gc

if not sys.warnoptions:
    warnings.simplefilter("ignore")

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')



# Hyperparameters of the model
input_dim =  28*28
output_dim = 10

# Number of neurons
N = 7000

# Epochs needed to train the model
epochs = 5

# Number of iterations performed by the train and test inference on every batch of data.
num_iterations = 10000#3
iterations_test = 5000
t_test = 5000

# Parameters of the algorithm. 
p_lr = 0.00005
x_lr = 0.5
xlr_test = 0.5
xl_discount = 0.9

# Name of the dataset
dataset_name = 'MNIST'
s = 1
wd = 0.5
acti_func = nn.ReLU()
partial = 3000
partial2  = 50
batch_size = 500 
test_batch_size = 2000

if 'MNIST' in dataset_name:
    fraction = 392
    input_shape = (-1,1,28,28)
if 'CIFAR' in dataset_name:
    fraction = 512
    input_shape = (-1,3,32,32)


#################################################################################

def reconstruct(model, dataloader, dataset_name,M):

    def prediction(t, pc_layer,pc_trainer):
        model[0].weight.data.copy_(model[0].weight.data * M.cuda())
        if 'CIFAR' in dataset_name:
            pc_layer.get_x().data[:,:fraction].copy_(lab_b)
            pc_layer.get_x().data[:,1024:1024+fraction].copy_(lab_r)
            pc_layer.get_x().data[:,2048:2048+fraction].copy_(lab_g)
        if 'MNIST' in dataset_name:
            pc_layer.get_x().data[:,:fraction].copy_(lab)
        if t == 0:
            prediction_trainer.inputs = pc_layer.get_x()
        
    callback_after_t_kwargs={'pc_layer':pc_layer.cuda(),'pc_trainer':pc_trainer}
    loss = 0.0
    for i, (images, labels) in enumerate(dataloader):
        model.train()
        num_rows = 10

        if 'CIFAR' in dataset_name:
            lab_b = images[:,:fraction]
            lab_r = images[:,1024:1024+fraction]
            lab_g = images[:,2048:2048+fraction]
        if 'MNIST' in dataset_name:
            lab = images[:,:fraction]

        if 'CIFAR' in dataset_name:
            zeros = torch.zeros((test_batch_size,3072))
            zeros[:,:fraction] = lab_b
            zeros[:,1024:1024+fraction] = lab_r
            zeros[:,2048:2048+fraction] = lab_g
        if 'MNIST' in dataset_name:
            zeros = torch.zeros((test_batch_size,784))
            zeros[:,:fraction] = lab

        images = Variable(images.view(-1,1).cuda())
        in_data=torch.zeros(test_batch_size, N).cuda()
        results = prediction_trainer.train_on_batch(in_data.float(),is_log_progress=False,callback_after_t=prediction,callback_after_t_kwargs=callback_after_t_kwargs)
        loss += results['overall'][len(results['overall'])-1]
        reconstruction = pc_layer.get_x().data.cpu().view(test_batch_size,-1)[:,:input_dim]
        noise = images.view(test_batch_size,-1).cpu() - reconstruction
        
        if partial2 > 0: 
            image = torch.cat((zeros.view(input_shape)[0::partial2], 
                                reconstruction.view(input_shape)[0::partial2],
                                images.view(input_shape)[0::partial2].cpu(),
                                noise.view(input_shape)[0::partial2]))
            save_image(image.cpu(), f"./assembly_results/reconstruction0727_newer/smaller_partial/-{dataset_name} - N={N},s={s},wd={wd},partial1={partial},partial2={partial2},batch_size={batch_size},frac={fraction},T={num_iterations},xlr={x_lr},plr={p_lr},T_Test={t_test},Epoch={epoch}.png", nrow=10)
        else:
            image = torch.cat((zeros.view(input_shape)[0:10],
                                reconstruction.view(input_shape)[0:10],
                                images.view(input_shape)[0:10].cpu(),
                                noise.view(input_shape)[0:10]))
            save_image(image.cpu(), f"./assembly_results/reconstruction0727_newer/smaller_partial/-{dataset_name} - N={N},s={s},wd={wd},partial1={partial},partial2={partial2},batch_size={batch_size},frac={fraction},T={num_iterations},xlr={x_lr},plr={p_lr},T_Test={t_test},Epoch={epoch}.png", nrow=10)
        break

###########################################################################################
partial2 = 5
N_list = [9000]#[9000, 13000, 17000， 24000,27000]
s_list = [1]#[0.8, 1]
T_list = [100]
wd_list = [1]
partial2_list = [5, 1]#5,10,25 #2,4, 5, 10
frac_list = [392] #int(1024/3),int(1024/4),int(1024/5)
partial_list = [1000, 3000, 1000]# 3000,5, 30, 300,  150, 400, 600, 70,150
grid = [(x, y, z,q) for x in N_list for y in T_list for z in partial_list for q in wd_list]

for N, num_iterations, partial, wd in grid:
    batch_size = 250#int(partial)
    test_batch_size = partial2*10
    train_loader, test_loader = u.load_dataset(dataset_name, partial, partial2, batch_size, test_batch_size, augmentation = False, shuffle = True)
    print('N={},s={},partial={},partial2={},batch_size={},fraction={}'.format(N, s, partial,partial2, batch_size,fraction))
    for seed in range(1):
        if s < 1:
            M = (torch.rand(size=(N,N)) < s).int()
        else:
            M = torch.ones((N,N))

        M[:input_dim,:input_dim] = torch.zeros((input_dim,input_dim))
        M[-output_dim:,-output_dim:] = torch.zeros((output_dim,output_dim))
        M.cuda()
        pc_layer = pc.PCLayer()
        model = nn.Sequential(nn.Linear(N, N),acti_func,pc_layer)

        model.train()
        model.cuda()

        pc_trainer = pc.PCTrainer(model,optimizer_x_kwargs={'lr': x_lr},optimizer_p_fn=optim.Adam, optimizer_p_kwargs={'lr': p_lr,'weight_decay':wd}, T = num_iterations,x_lr_discount=xl_discount,plot_progress_at=[],early_stop_condition='True')
        pc_predictor = pc.PCTrainer(model, optimizer_x_kwargs={'lr': xlr_test}, optimizer_p_kwargs={'lr': 0},optimizer_p_fn=optim.SGD, T = iterations_test,x_lr_discount=xl_discount,plot_progress_at=[],early_stop_condition='True')

        def callback_after_t(t, pc_layer, data, target, pc_trainer):
            model[0].weight.data.copy_(model[0].weight.data * M.cuda())
            pc_layer.get_x().data[:, :input_dim].copy_(data)
            pc_layer.get_x().data[:, -output_dim:].copy_(target)
            if t == 0:
                pc_trainer.inputs = pc_layer.get_x()

        for epoch in range(epochs):
            print("This is epoch ", epoch)
            print("Training starts...")
            for i, (data, target) in enumerate(train_loader):
                print(i)
                in_data=torch.zeros(batch_size, N).cuda()
                callback_after_t_kwargs={'pc_layer':pc_layer.cuda(),'data':data.cuda(),'target':target.cuda(),'pc_trainer':pc_trainer}
                pc_trainer.train_on_batch(inputs=in_data, is_log_progress=False,callback_after_t=callback_after_t,callback_after_t_kwargs=callback_after_t_kwargs)
            print("Prediction starts...")
            if epoch % 1 == 0:
                prediction_trainer = pc.PCTrainer(model, optimizer_x_kwargs={'lr': xlr_test}, optimizer_p_kwargs={'lr': 0},optimizer_p_fn=optim.SGD,x_lr_discount=0.9, T = t_test,early_stop_condition='True')
                reconstruct(model, test_loader,dataset_name,M)
        del pc_trainer
        del prediction_trainer
        del pc_predictor
        gc.collect()
        torch.cuda.empty_cache()